@@ -30,4 +30,8 @@ struct stmt *meta_stmt_meta_iiftype(const struct location *loc, uint16_t type);
 
 const struct datatype ifindex_type;
 
+struct error_record *meta_key_parse(const struct location *loc,
+				    const char *name,
+				    unsigned int *value);
+
 #endif /* NFTABLES_META_H */
