@@ -25,6 +25,7 @@
 #include <netlink.h>
 #include <erec.h>
 #include <mnl.h>
+#include <iface.h>
 #include <cli.h>
 
 unsigned int max_errors = 10;
@@ -106,19 +107,18 @@ static void show_help(const char *name)
 "Usage: %s [ options ] [ cmds... ]\n"
 "\n"
 "Options:\n"
-"  -h/--help			Show this help\n"
-"  -v/--version			Show version information\n"
+"  -h, --help			Show this help\n"
+"  -v, --version			Show version information\n"
 "\n"
-"  -f/--file <filename>		Read input from <filename>\n"
-"  -i/--interactive		Read input from interactive CLI\n"
+"  -f, --file <filename>		Read input from <filename>\n"
+"  -i, --interactive		Read input from interactive CLI\n"
 "\n"
-"  -n/--numeric			When specified once, show network addresses numerically (default behaviour).\n"
-"  				When specified twice, show Internet services,\n"
-"				user IDs and group IDs numerically.\n"
-"				When specified thrice, also show protocols numerically.\n"
+"  -n, --numeric			When specified once, show network addresses numerically (default behaviour).\n"
+"  				Specify twice to also show Internet services (port numbers) numerically.\n"
+"				Specify three times to also show protocols, user IDs, and group IDs numerically.\n"
 "  -N				Translate IP addresses to names.\n"
-"  -a/--handle			Output rule handle.\n"
-"  -I/--includepath <directory>	Add <directory> to the paths searched for include files.\n"
+"  -a, --handle			Output rule handle.\n"
+"  -I, --includepath <directory>	Add <directory> to the paths searched for include files.\n"
 #ifdef DEBUG
 "  --debug <level [,level...]>	Specify debugging level (scanner, parser, eval, netlink, mnl, proto-ctx, segtree, all)\n"
 #endif
@@ -181,7 +181,6 @@ static int nft_netlink(struct parser_state *state, struct list_head *msgs)
 	bool batch_supported = netlink_batch_supported();
 	int ret = 0;
 
-	netlink_genid_get();
 	mnl_batch_init();
 
 	batch_seqnum = mnl_batch_begin();
@@ -229,15 +228,12 @@ int nft_run(void *scanner, struct parser_state *state, struct list_head *msgs)
 	int ret;
 
 	ret = nft_parse(scanner, state);
-	if (ret != 0 || state->nerrs > 0)
-		return -1;
-retry:
-	ret = nft_netlink(state, msgs);
-	if (ret < 0 && errno == EINTR) {
-		netlink_restart();
-		goto retry;
+	if (ret != 0 || state->nerrs > 0) {
+		ret = -1;
+		goto err1;
 	}
-
+	ret = nft_netlink(state, msgs);
+err1:
 	list_for_each_entry_safe(cmd, next, &state->cmds, list) {
 		list_del(&cmd->list);
 		cmd_free(cmd);
@@ -331,16 +327,21 @@ int main(int argc, char * const *argv)
 		for (len = 0, i = optind; i < argc; i++)
 			len += strlen(argv[i]) + strlen(" ");
 
-		buf = xzalloc(len + 1);
+		buf = xzalloc(len + 2);
 		for (i = optind; i < argc; i++) {
 			strcat(buf, argv[i]);
 			if (i + 1 < argc)
 				strcat(buf, " ");
 		}
+		strcat(buf, "\n");
 		parser_init(&state, &msgs);
 		scanner = scanner_init(&state);
 		scanner_push_buffer(scanner, &indesc_cmdline, buf);
 	} else if (filename != NULL) {
+		rc = cache_update(CMD_INVALID, &msgs);
+		if (rc < 0)
+			return rc;
+
 		parser_init(&state, &msgs);
 		scanner = scanner_init(&state);
 		if (scanner_read_file(scanner, filename, &internal_location) < 0)
@@ -362,7 +363,9 @@ int main(int argc, char * const *argv)
 out:
 	scanner_destroy(scanner);
 	erec_print_list(stderr, &msgs);
-
 	xfree(buf);
+	cache_release();
+	iface_cache_release();
+
 	return rc;
 }
